"""Tools for safely saving and loading pickles using an AllowedList"""

import copyreg
import io
import pickle
import typing as T

import warnings

from snowfakery.object_rows import NicknameSlot, ObjectReference, ObjectRow

_DISPATCH_TABLE = copyreg.dispatch_table.copy()
_DISPATCH_TABLE[NicknameSlot] = lambda n: (
    ObjectReference,
    (n._tablename, n.allocated_id),
)
_DISPATCH_TABLE[ObjectRow] = lambda v: (
    ObjectRow,
    (v._tablename, v._values),
)


def restricted_dumps(data):
    """Only allow saving "safe" classes"""
    outs = io.BytesIO()
    pickler = pickle.Pickler(outs)
    pickler.dispatch_table = _DISPATCH_TABLE
    pickler.dump(data)
    return outs.getvalue()


class Type_Cannot_Be_Used_With_Random_Reference(T.NamedTuple):
    """This type cannot be unpickled."""

    module: str
    name: str


_SAFE_CLASSES = {
    ("snowfakery.object_rows", "ObjectRow"),
    ("snowfakery.object_rows", "ObjectReference"),
    ("snowfakery.object_rows", "LazyLoadedObjectReference"),
    ("decimal", "Decimal"),
    ("datetime", "date"),
    ("datetime", "datetime"),
    ("datetime", "timedelta"),
    ("datetime", "timezone"),
}


class RestrictedUnpickler(pickle.Unpickler):
    """Safe unpickler with an allowed-list"""

    count = 0

    def find_class(self, module, name):
        # Only allow safe classes from builtins.
        if (module, name) in _SAFE_CLASSES:
            return super().find_class(module, name)
        else:
            # Return a "safe" object that does nothing.
            if RestrictedUnpickler.count < 10:
                warnings.warn(f"Cannot save and refer to {module}, {name}")
                RestrictedUnpickler.count += 1
            return lambda *args: Type_Cannot_Be_Used_With_Random_Reference(module, name)


def restricted_loads(s):
    """Helper function analogous to pickle.loads()."""
    return RestrictedUnpickler(io.BytesIO(s)).load()



import typing as T
from contextlib import contextmanager
from pathlib import Path

from click.utils import LazyFile

OpenFileLike = T.Union[T.TextIO, LazyFile]
FileLike = T.Union[OpenFileLike, Path, str]


@contextmanager
def open_file_like(
    file_like: T.Optional[FileLike], mode, **kwargs
) -> T.ContextManager[T.Tuple[str, OpenFileLike]]:
    """Look at a file-like or path-like object and open it

    Returns a) the path it was given OR a best-guess of the path and
            b) the open file object

    Used properly as a context manager, it will close the file later
    if appropriate.
    """
    if not file_like:
        yield None, None
    if isinstance(file_like, str):
        file_like = Path(file_like)

    if isinstance(file_like, Path):
        with file_like.open(mode, **kwargs) as f:
            yield file_like, f

    elif hasattr(file_like, "name"):
        yield file_like.name, file_like

    elif hasattr(file_like, "read"):
        yield None, file_like



import re
import setuptools

from typing import List


def parse_requirements_file(requirements_file) -> List[str]:
    requirements = []
    for req in requirements_file.read().splitlines():
        # skip comments and hash lines
        if not (re.match(r"\s*#", req) or re.match(r"\s*--hash", req)):
            req = req.split(" ")[0]
            requirements.append(req)
    return requirements


with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements/prod.txt") as requirements_file:
    requirements = parse_requirements_file(requirements_file)

with open("requirements/dev.txt") as dev_requirements_file:
    dev_requirements = parse_requirements_file(dev_requirements_file)

# get the version into a global variable named "version"
with open("snowfakery/version.txt") as f:
    version = f.read().strip()

packages = [
    p
    for p in setuptools.find_namespace_packages()
    if p.startswith("snowfakery") and not p.startswith("snowfakery.docs")
]

setuptools.setup(
    name="snowfakery",
    version=version,
    author="Paul Prescod",
    author_email="pprescod@salesforce.com",
    description=(
        "Snowfakery is a tool for generating fake data that has "
        "relations between tables. Every row is faked data, but also "
        "unique and random, like a snowflake."
    ),
    entry_points={
        "console_scripts": [
            "snowfakery=snowfakery.cli:main",
            "snowbench=snowfakery.tools.snowbench:main",
        ],
        "mkdocs.plugins": ["snowfakery_fakes=snowfakery.tools.mkdocs:Plugin"],
    },
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SFDO-Tooling/Snowfakery",
    packages=packages,
    package_dir={"snowfakery": "snowfakery"},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=requirements,
    tests_require=dev_requirements,
    data_files=[("requirements", ["requirements/prod.txt", "requirements/dev.txt"])],
)



""" Benchmarking tool for the RowHistory backing database"""

from snowfakery.row_history import RowHistory
from snowfakery.object_rows import RowHistoryCV
from time import time

COUNT = 200_000


def main(count):
    rh: RowHistory = RowHistory()
    RowHistoryCV.set(rh)
    start = time()
    for i in range(0, COUNT, 4):
        rh.save_row(
            "table",
            None,
            {
                "id": i,
                "blah": "blah",
            },
        )

        rh.save_row(
            "table",
            "Nicknamed",
            {
                "id": i + 1,
                "blah": "blah",
            },
        )

        rh.save_row(
            "table",
            "Nicknamed2",
            {
                "id": i + 2,
                "blah": "blah",
            },
        )

        if i % 1000 == 0:
            nickname = "Sparse"
        else:
            nickname = "Unused"

        rh.save_row(
            "table",
            nickname,
            {
                "id": i + 3,
                "blah": "blah",
            },
        )

    end = time()
    print("Saved", end - start)
    start = time()
    for i in range(0, COUNT):
        rh.load_row("table", i)
        assert rh.random_row_reference("table", "current-iteration", False).blah
        assert rh.random_row_reference("Nicknamed", "current-iteration", False).blah
        assert rh.random_row_reference("Nicknamed2", "current-iteration", False).blah
        assert rh.random_row_reference("Sparse", "current-iteration", False).blah
    end = time()
    print("Loaded", end - start)


main(COUNT)
